###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

"""
Author: Ariane Mora
Date: September 2024
"""
import typer
import sys
import pandas as pd
from oligopoolio import *
from oligopoolio.oligos import *
import os
from typing_extensions import Annotated
from os.path import dirname, join as joinpath

app = typer.Typer()

def fasta_to_df(fasta):
    rows = []
    records = list(SeqIO.parse(fasta, "fasta"))
    done_records = []
    # Remove all the ids
    for record in records:
        new_id = re.sub('[^0-9a-zA-Z]+', '', record.id)
        if new_id not in done_records:
            rows.append([new_id, record.seq])
        else:
            u.warn_p(['Had a duplicate record! Only keeping the first entry, duplicate ID:', record.id])
    df = pd.DataFrame(rows, columns=['id', 'seq'])
    return df
@app.command()
def run(input_file: Annotated[str, typer.Argument(help="Full path to query fasta or csv (note have simple IDs "\
                                                     "otherwise we'll remove all funky characters.). If you pass as a csv you need to also supply the ")],
        prime5f_overlap: Annotated[str, typer.Argument(help="Overlap of the 5prime region for the backbone.")],
        prime3f_overlap: Annotated[str, typer.Argument(help="Overlap of the 3prime region for the backbone "
                                                            "(same strand as 5prime).")],
        seq_start: Annotated[str, typer.Argument(help="Codons to add to the start of the sequence")] = "",
        seq_end: Annotated[str, typer.Option(help="End of the sequence")] = "",
        oligo_start: Annotated[str, typer.Argument(help="Codons to add to the start of each oligo e.g. for amplification.")] = "",
        oligo_end: Annotated[str, typer.Argument(help="Codons to add to the end of each oligo e.g. for amplification.")] = "",
        min_oligo_len: Annotated[int, typer.Option(help="Minimum length of the oligo")] = 100,
        max_oligo_len: Annotated[int, typer.Option(help="Maximum length of the oligo")] = 200,
        optimal_tm: Annotated[int, typer.Option(help="Optimal temperature of the overlap between oligos (C).")] = 62,
        genbank_file: Annotated[int, typer.Option(help="Genbank file in which to add the oligos for visualisation")] = 62,
        output_folder: Annotated[str, typer.Option(help="Where to store results (full path!)")] = 'Current Directory',
        run_name: Annotated[str, typer.Option(help="Name of the run, default is oligos")] = 'oligos',
        id_col: Annotated[str, typer.Option(help="id column in df if df passed (csv) rather than fasta")] = 'id',
        seq_col: Annotated[str, typer.Option( help="codon optimized sequence column.)")] = 'seq',
        ):
    """
    Find similar proteins based on sequence or structural identity in order to annotate these using
    BLAST and FoldSeek. Also annotate with ProteInfer and CLEAN.
    """
    output_folder = output_folder if output_folder != 'Current Directory' else os.getcwd()
    if '.fa' in input_file or '.fasta' in input_file and '.csv' not in input_file:
        df = fasta_to_df(input_file)
    else:
        df = pd.read_csv(input_file)

    # Run the pipeline
    get_oligos(run_name, df, id_col, seq_col, output_folder,
             database, clean_dir=clean_dir, proteinfer_dir=proteinfer_dir,
             run_method=run_method, keep_dups=keep_dups, args_blast=args_blast, args_foldseek=args_foldseek,
             args_proteinfer=args_proteinfer, args_clean=args_clean, methods=methods, foldseek_db=foldseek_db)


if __name__ == "__main__":
    app()

# Example command
# oligopoolio input_df.csv Uniprot_reviewed_catalytic_activity_06032025.fasta --methods blast --output-folder output/ --run-name omgprot50


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from oligopoolio.oligopools import *
from oligopoolio.primers import *
from oligopoolio.oligos import *

import unittest
from sciutil import *

u = SciUtil()

test_data_dir = os.path.join(os.path.dirname(__file__), 'data/')


class TestPrimers(unittest.TestCase):

    def test_primer_making(self):
        # Placeholder gene ends (replace with your actual gene sequences)
        gene = "ATGAGCGATCTGCATAACGAGTCCATTTTTATTACCGGCGGCGGATCGGGATTAGGGCTGGCGCTGGTCGAGCGATTTAT\
        CGAAGAAGGCGCGCAGGTTGCCACGCTGGAACTGTCGGCGGCAAAAGTCGCCAGTCTGCGTCAGCGATTTGGCGAACATA\
        TTCTGGCGGTGGAAGGTAACGTGACCTGTTATGCCGATTATCAACGCGCGGTCGATCAGATCCTGACTCGTTCCGGCAAG\
        CTGGATTGTTTTATCGGCAATGCAGGCATCTGGGATCACAATGCCTCACTGGTTAATACTCCCGCAGAGACGCTCGAAAC\
        CGGCTTCCACGAGCTGTTTAACGTCAACGTACTCGGTTACCTGCTGGGCGCAAAAGCCTGCGCTCCGGCGTTAATCGCCA\
        GTGAAGGCAGCATGATTTTCACACTGTCAAATGCCGCCTGGTATCCTGGCGGCGGTGGCCCGCTGTACACCGCCAGTAAA\
        CATGCCGCAACCGGACTTATTCGCCAACTGGCTTATGAACTGGCACCGAAAGTGCGGGTGAATGGCGTCGGCCCGTGTGG\
        TATGGCCAGCGACCTGCGCGGCCCACAGGCGCTCGGGCAAAGTGAAACCTCGATAATGCAGTCTCTGACGCCGGAGAAAA\
        TTGCCGCCATTTTACCGCTGCAATTTTTCCCGCAACCGGCGGATTTTACGGGGCCGTATGTGATGTTGGCATCGCGGCGC\
        AATAATCGCGCATTAAGCGGTGTGATGATCAACGCTGATGCGGGTTTAGCGATTCGCGGCATTCGCCACGTAGCGGCTGG\
        GCTGGATCTTTAA"

        # Standard pET-22b(+) primer sequences
        forward_plasmid_primer = "GGAGATATACATATG"
        reverse_plasmid_primer = "GCTTTGTTAGCAGCCGGATCTCA"

        # Desired Tm range for optimization
        desired_tm = 62.0  # Target melting temperature in °C
        tm_tolerance = 5.0  # Allowable deviation from the desired Tm

        # Generate and optimize forward primer
        min_length = 13
        max_length = 20
        forward_gene_primer, forward_tm = optimize_primer(forward_plasmid_primer, gene, desired_tm, 'forward',
                                                          min_length, max_length, tm_tolerance)
        reverse_gene_primer, reverse_tm = optimize_primer(reverse_plasmid_primer, gene, desired_tm, 'reverse',
                                                          min_length, max_length, tm_tolerance)

        print(f"Forward Gene Primer: 5'-{forward_gene_primer}-3' (Tm: {forward_tm:.2f} °C)")
        print(f"Reverse Gene Primer: 5'-{reverse_gene_primer}-3' (Tm: {reverse_tm:.2f} °C)")

    def test_get_flanking(self):
        gff_file = f'{test_data_dir}genome_NEB_B/genomic.gff'
        reference_fasta = f'{test_data_dir}genome_NEB_B/GCF_001559615.2_ASM155961v2_genomic.fna'
        gene_name = 'udp'
        seqid, start, end, strand, upstream_flank, downstream_flank, gene_seq = get_flanking_primers(gene_name,
                                                                                                     gff_file,
                                                                                                     reference_fasta)

        assert gene_seq[:3] == 'ATG'  # i.e. methionine
        assert gene_seq[-3:] == 'TAA'
        assert len(upstream_flank) == 50
        assert len(downstream_flank) == 50
        assert 'AAAAC' == downstream_flank[:5]
        assert 'AAGT' == upstream_flank[:4]

    def test_get_flanking_reverse(self):
        gff_file = f'{test_data_dir}genome_NEB_B/genomic.gff'
        reference_fasta = f'{test_data_dir}genome_NEB_B/GCF_001559615.2_ASM155961v2_genomic.fna'
        gene_name = 'ysgA'
        seqid, start, end, strand, upstream_flank, downstream_flank, gene_seq = get_flanking_primers(gene_name,
                                                                                                     gff_file,
                                                                                                     reference_fasta)

        actual_gene_seq = 'ATGGCAACAACACAACAATCTGGATTTGCACCTGC\
            TGCATCGCCTCTCGCTTCGACCATCGTTCAGACTCCGGACGACGC\
            GATTGTGGCGGGCTTCACCTCTATCCCTTCACAAGGGGATAACATGCCTGCTTACCATGCCAGACCAAAGCAAAGCGATG\
            GCCCACTGCCAGTGGTCATTGTAGTGCAGGAAATTTTTGGCGTGCATGAACATATCCGCGATATTTGTCGCCGTCTGGCG\
            CTGGAGGGGTATCTGGCTATCGCACCTGAACTTTACTTCCGCGAAGGCGATCCGAATGATTTTGCCGATATCCCTACGCT\
            GCTTAGCGGTCTGGTAGCAAAAGTGCCTGACTCGCAGGTGCTGGCCGATCTCGATCATGTCGCCAGTTGGGCGTCACGCA\
            ACGGCGGCGATGTTCATCGTTTAATGATCACCGGATTCTGCTGGGGTGGACGTATCACCTGGCTGTATGCCGCGCATAAT\
            CCACAGCTAAAAGCCGCAGTGGCGTGGTACGGCAAGCTGACAGGCGATAAGTCGCTGAATTCACCGAAACAACCTGTTGA\
            TATCGCAACCGATCTTAACGCGCCGGTTCTCGGCTTATATGGCGGCCAGGATAACAGCATTCCGCAAGAGAGCGTGGAAA\
            CGATGCGCCAGGCGCTGCGGGCAGCAAACGCGAAAGCAGAGATTATCGTCTACCCGGATGCCGGGCATGCATTCAACGCC\
            GATTATCGCCCGAGCTATCATGCCGAGTCTGCGAAAGACGGCTGGCAGCGAATGTTGGAATGGTTTACACAGTATGGTGTTAAAAAGTAA'

        print(gene_seq[:20])
        print(actual_gene_seq[:20])
        print(gene_seq[-20:])
        print(actual_gene_seq[-20:])
        assert gene_seq[:20] == actual_gene_seq[:20]
        assert gene_seq[-20:] == actual_gene_seq[-20:]

        assert gene_seq[:3] == 'ATG'  # i.e. methionine
        assert gene_seq[-3:] == 'TAA'
        assert len(upstream_flank) == 50
        print(downstream_flank, upstream_flank)
        assert len(downstream_flank) == 50
        assert 'TACC' == upstream_flank[-4:]
        assert 'CGCC' == downstream_flank[:4]
        # At the moment I'm unsure if the flanks need to be reversed?

    def test_from_fasta(self):
        fasta_file = f'{test_data_dir}example_fasta.fasta'
        df = make_primers_IDT(fasta_file, remove_stop_codon=True, his_tag='',
                              max_length=60, min_length=15, tm_tolerance=30, desired_tm=62.0,
                              forward_primer='gaaataattttgtttaactttaagaaggagatatacat',
                              reverse_primer='ctttgttagcagccggatc')
        assert len(df) == 8
        assert df['Sequence'].values[0] == 'ctttaagaaggagatatacatATGACCATAGACAAAAATTGG'

    def test_single_oligo(self):
        fasta_file = f'{test_data_dir}example_fasta.fasta'
        df = make_oligo_single(fasta_file, forward_primer='gaaataattttgtttaactttaagaaggagatatacat',
                          forward_primer_len=15, reverse_primer='gatccggctgctaacaaag', reverse_primer_len=15,
                          max_len=320)
        assert df['forward_primer'].values[0] == 'gaaggagatatacat'


    def test_double_oligo(self):
        fasta_file = f'{test_data_dir}example_fasta.fasta'

        df = make_oligo_double(fasta_file, forward_primer='gaaataattttgtttaactttaagaaggagatatacat',
                          forward_primer_len=15, reverse_primer='gatccggctgctaacaaag', reverse_primer_len=15,
                          max_len=640,
                          overlap_len=9)

    def test_double_oligo_run(self):
        fasta_file = f'/Users/arianemora/Documents/code/degradeo/manuscript/data/oligopool_random_forest.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_oligo_double(fasta_file, forward_primer='gaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=15, reverse_primer='gatccggctgctaacaaag', reverse_primer_len=15,
                               max_len=670,
                               overlap_len=24)
        df.to_csv('/Users/arianemora/Documents/code/degradeo/manuscript/data/order_oligopool_random_forest_NT_SwissProt_prediction_DEHP_25092024.csv', index=False)

    def test_double_oligo_run_v2(self):
        # Read in a csv and then convert to fasta
        df = pd.read_csv('~/Documents/oligo_seq_order_metagenomics.csv')
        # save the seqs to a fasta file here
        with open('oligo_seq_order_metagenomics.fasta', 'w+') as fout:
            for seq_name, seq in df[['sequence_name', 'codon_optiimized']].values:
                fout.write(f'>{seq_name}\n{seq}\n')
        fasta_file = f'oligo_seq_order_metagenomics.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_oligo_double(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=46, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=24)
        df.to_csv('oligo_seq_order_metagenomics_order.csv', index=False)

    def test_double_oligo_run_v3(self):
        # # Read in a csv and then convert to fasta
        # df = pd.read_csv('~/Documents/oligo_seq_order_metagenomics.csv')
        # # save the seqs to a fasta file here
        # with open('oligo_seq_order_metagenomics.fasta', 'w+') as fout:
        #     for seq_name, seq in df[['sequence_name', 'codon_optiimized']].values:
        #         fout.write(f'>{seq_name}\n{seq}\n')
        fasta_file = f'oligo_seq_order_metagenomics.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_oligo_double(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=46, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=24)
        df.to_csv('oligo_seq_order_metagenomics_order_rev_comp_2.csv', index=False)

    def test_double_oligo_TWIST_thermofiles(self):
        # # Read in a csv and then convert to fasta
        df = pd.read_csv('~/Documents/optimized_thermofiles.csv')
        # save the seqs to a fasta file here
        with open('optimized_thermofiles.fasta', 'w+') as fout:
            for seq_name, seq in df[['Name', 'Optimized Sequence']].values:
                print(len(seq))
                fout.write(f'>{seq_name}\n{seq}\n')
        fasta_file = f'optimized_thermofiles.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_oligo_double(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=46, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=900,
                               overlap_len=24)
        df.to_csv('oligo_seq_order_optimized_thermofiles_order_rev_comp_2.csv', index=False)

    def test_double_oligo_TWIST(self):
        fasta_file = f'oligo_seq_order_metagenomics.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_oligo_double(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=38, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=18)
        df.to_csv('oligo_seq_order_metagenomics_double.csv', index=False)
    def test_double_oligo_many(self):
        fasta_file = f'oligo_seq_order_metagenomics.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_splitty_oligo(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=38, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=18)
        df.to_csv('oligo_seq_order_metagenomics_splitty.csv', index=False)

    def test_multi_oligo_rxnfp_chai(self):
        df = pd.read_csv('../predictions_rxnfp_chai/all-sequences.csv')
        # save the seqs to a fasta file here
        with open('../predictions_rxnfp_chai/optimized_rxnfp.fasta', 'w+') as fout:
            for seq_name, seq in df[['Name', 'Full Optimized Sequence']].values:
                fout.write(f'>{seq_name}\n{seq}\n')
        fasta_file = '../predictions_rxnfp_chai/optimized_rxnfp.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_splitty_oligo(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=38, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=18)
        df.to_csv('../predictions_rxnfp_chai/oligo_seq_order_rxnfp_splitty.csv', index=False)


    def test_multi_oligo_rxnfp_chai_thermo(self):
        df = pd.read_csv('../predictions_rxnfp_chai/codon_optimized_creep_predictions_thermofiles/optimized.csv')
        # save the seqs to a fasta file here
        with open('../predictions_rxnfp_chai/codon_optimized_creep_predictions_thermofiles/optimized_chemberta.fasta', 'w+') as fout:
            for seq_name, seq in df[['Name', 'Optimized Sequence']].values:
                fout.write(f'>{seq_name}\n{seq}\n')
        fasta_file = '../predictions_rxnfp_chai/codon_optimized_creep_predictions_thermofiles/optimized_chemberta.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_splitty_oligo(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat', # CTCGAGCACCACCACCACCACCACTGA
                               forward_primer_len=38, reverse_primer='gatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=18)
        df.to_csv('../predictions_rxnfp_chai/codon_optimized_creep_predictions_thermofiles_splitty.csv', index=False)


    def test_double_oligo_rxnfp_chai(self):
        df = pd.read_csv('../predictions_rxnfp_chai/all-sequences.csv')
        # save the seqs to a fasta file here
        with open('../predictions_rxnfp_chai/optimized_rxnfp_noTAA.fasta', 'w+') as fout:
            for seq_name, seq in df[['Name', 'Optimized Sequence']].values:
                fout.write(f'>{seq_name}\n{seq}\n')
        fasta_file = '../predictions_rxnfp_chai/optimized_rxnfp_noTAA.fasta' #oligopool_random_forest_NT_SwissProt_ActiveSite_prediction_DEHP_25092024.txt'

        df = make_oligo_double(fasta_file, forward_primer='cccctctagaaataattttgtttaactttaagaaggagatatacat',
                               forward_primer_len=38, reverse_primer='CTCGAGCACCACCACCACCACCACTGAgatccggctgctaacaaag',
                               reverse_primer_len=46,
                               max_len=670,
                               overlap_len=18)
        df.to_csv('../predictions_rxnfp_chai/oligo_seq_order_rxnfp_double.csv', index=False)


    def test_long_oligos(self):
        df = pd.read_csv('../scripts/degradeo/predictions_PBP_12042025/pbp_codon_optimized_no_spaces_5.csv')

        min_gc = 0.25
        max_gc = 0.65
        min_tm = 10
        max_tm = 1000
        min_segment_length = 200
        max_segment_length = 250
        max_length = 500
        primer_lower = 'GATCCGGC'.lower()

        output_directory = 'output_long/'
        oligo_df = get_oligos(df, 'Optimized Sequence', 'Name', output_directory,
                              'gaaataattttgtttaactttaagaaggagatatacat',
                              primer_lower, sequence_end='CTCGAGCACCACCACCACCACCACTGA',
                              min_gc=min_gc, max_gc=max_gc, min_tm=min_tm, max_tm=max_tm,
                              min_segment_length=min_segment_length,
                              max_segment_length=max_segment_length,
                              genbank_file="../scripts/degradeo/predictions_PBP_12042025/base-pet22b-base-anm.gb", insert_position=5193, simple=True,
                              codon_optimize=False)




###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

"""
Author: Ariane Mora
Date: September 2024
"""
__title__ = 'oligopoolio'
__description__ = 'pools and primers mateo'
__url__ = 'https://github.com/arianemora/oligopoolio/'
__version__ = '0.0.7'
__author__ = 'Ariane Mora'
__author_email__ = 'ariane.n.mora@gmail.com'
__license__ = 'GPL3'

from oligopoolio.primers import *
from oligopoolio.oligopools import *
from oligopoolio.oligos import *



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

"""
Author: Ariane Mora
Date: 24th March 2024
"""
import random
import string
from tqdm import tqdm

from Bio.Seq import Seq
import primer3
from sciutil import SciUtil
import gffutils
import pandas as pd
from Bio import SeqIO
import pandas as pd

u = SciUtil()


def reverse(seq):
    """ Just reverse a sequence. """
    return reversed(seq)


def check_aa(seq, allow_gaps=False):
    """ Check that the sequence starts with a methionine and ends with a stop codon."""
    bases = 'ACDEFGHIKLMNPQRSTVWY'
    if seq[0] != 'M':
        u.warn_p(['Your sequence did not start with Methionine...'])
        return False
    if 'M' in seq[1:]:
        u.warn_p(['Your sequence had multiple Methionines...'])
        return False
    for b in seq:
        if allow_gaps:
            bases = 'ACDEFGHIKLMNPQRSTVWY-'
        if b not in bases:
            u.warn_p(['Your sequence contained non-canonical bases...', b])
            return False
    return True


def check_nt(seq, allow_gaps = False):
    """ Check that the sequence starts with a methionine and ends with a stop codon."""
    bases = 'ATGC'
    if seq[:3] != 'ATG':
        u.warn_p(['Your sequence did not start with Methionine ATG...'])
        return False
    if seq[-3:] != 'TAA':
        u.warn_p(['Your sequence did not end with stop codon TAA...'])
        return False
    if 'TAA' in seq[:-3]:
        u.warn_p(['Your sequence had multiple stop codons...'])
        return False
    if 'ATG' in seq[3:]:
        u.warn_p(['Your sequence had multiple start codons...'])
        return False
    for b in seq:
        if allow_gaps:
            bases = 'ATGC-'
        if b not in bases:
            u.warn_p(['Your sequence contained non-canonical bases...', b])
            return False
    return True


def clean_aa_seq(seq, allow_gaps=True, remove=False, replace='N', amino_acids=None):
    """
        Remove any non-canonical bases and swap them with N. amino_acids can be
        specified if the user wants non-canonical amino acid bases.
    """
    if allow_gaps:
        bases = amino_acids or list('ACDEFGHIKLMNPQRSTVWY-')  # Allow gaps
    else:
        bases = amino_acids or list('ACDEFGHIKLMNPQRSTVWY')
    seq = seq.strip().replace(' ', '')  # Remove any spaces
    if remove:
        replace = ''  # Swap the other base of N for -
    return ''.join([s if s in bases else replace for s in seq])


def clean_nt_seq(seq, allow_gaps=True, remove=False, replace='N'):
    """ Remove any non-canonical bases and swap them with N."""
    if allow_gaps:
        bases = ['A', 'T', 'G', 'C', '-']  # Allow gaps
    else:
        bases = ['A', 'T', 'G', 'C']
    seq = seq.strip().replace(' ', '')  # Remove any spaces
    if remove:
        replace = ''  # Swap the other base of N for -
    return ''.join([s if s in bases else replace for s in seq])


def reverse_complement(seq):
    """Generate the reverse complement of a DNA sequence."""
    seq = seq.strip().replace(' ', '') # Do a little cleaning
    complement = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
    return ''.join([complement[base] for base in reversed(seq)])


def gff_to_dataframe(db):
    # Function to convert GFF database to pandas DataFrame
    records = []

    # Iterate over each feature in the database
    for feature in db.all_features():
        record = {
            'seqid': feature.seqid,
            'source': feature.source,
            'feature_type': feature.featuretype,
            'start': feature.start,
            'end': feature.end,
            'score': feature.score,
            'strand': feature.strand,
            'phase': feature.frame,
        }

        # Add attributes as separate columns
        for key, value in feature.attributes.items():
            record[key] = value[0] if len(value) == 1 else value
        # Let's also do this for each of the dbx references making it a new column
        db_entries = record.get('Dbxref')
        if db_entries:
            if isinstance(db_entries, str):
                record[db_entries.split(':')[0]] = db_entries.split(':')[1]
            elif isinstance(db_entries, list):
                for entry in db_entries:
                    record[entry.split(':')[0]] = entry.split(':')[1]

        records.append(record)

    return pd.DataFrame(records)


def create_db_from_gff(gff_file):
    # Function to create a database from the GFF file
    db_file = gff_file + '.db'
    return gffutils.create_db(gff_file, dbfn=db_file, force=True, keep_order=False, merge_strategy='merge',
                              sort_attribute_values=True)


def get_flanking_primers(gene_id, gff_file, fasta_reference, upstream_flank_number=50, downstream_flank_number=50):
    """
    Get the flanking primers for a given gene from a reference sequence.
https://benchling.com/arnold_lab/f/lib_yyMnf2lS-trpb_landscape/prt_kpSpRW0e-1-red-mediated-e-coli-knockout/edit
    Note we expect the file format to be in that from NCBI.
    """
    # First get the gene location from the gff file.
    db = create_db_from_gff(gff_file)

    # Convert the database to a pandas DataFrame
    gff_df = gff_to_dataframe(db)
    gff_df.to_csv('test.csv')
    # Then get the position in the fasta file and ensure that the upstream starts at TAG and ends with TAA
    # i.e. start and stop codon.
    gene_df = gff_df[gff_df['Name'] == gene_id]
    start = gene_df['start'].values[0]
    end = gene_df['end'].values[0]
    strand = gene_df['strand'].values[0]
    seqid = gene_df['seqid'].values[0]
    # Get this from the fasta file now
    # There is probably a nicer way to do this with the sequence package but given I'm on a plane gonna do this dumb
    # dumb way and I'm really tired
    upstream_flank = ''
    gene_seq = ''
    downstream_flank = ''
    with open(fasta_reference, 'r+') as fin:
        seq_found = False
        sequence = ''
        for line in fin:
            if line[0] == '>':
                if seqid in line[1:].strip():  # This is our sequence so we just go to that position
                    # Need to get the next line
                    seq_found = True
                elif seq_found:
                    # This is now our sequnece and it is complete
                    if strand == '+':  # This is normal direction
                        gene_seq = sequence[start - 1:end]
                        upstream_flank = sequence[start - 1 - upstream_flank_number: start - 1]
                        downstream_flank = sequence[end: end + downstream_flank_number + 1]
                        return seqid, start, end, strand, upstream_flank, downstream_flank, gene_seq

                    else:  # reverse stranded so the upstream is actually from the end
                        gene_seq = sequence[start - (upstream_flank_number + 1):end + downstream_flank_number]
                        # Reverse complement this
                        gene_seq = reverse_complement(gene_seq)
                        upstream_flank = gene_seq[:upstream_flank_number]
                        downstream_flank = gene_seq[-downstream_flank_number:]
                        gene_seq = gene_seq[upstream_flank_number: -downstream_flank_number]
                        return seqid, start, end, strand, upstream_flank, downstream_flank, gene_seq
            elif seq_found:
                sequence += line.strip() # Build the sequence.
    if seq_found:
        # This is now our sequnece and it is complete
        if strand == '+':  # This is normal direction
            gene_seq = sequence[start - 1:end]
            upstream_flank = sequence[start - 1 - upstream_flank_number: start - 1]
            downstream_flank = sequence[end: end + downstream_flank_number]
        else:  # reverse stranded so the upstream is actually from the end
            # Possibly also reverset complement this???
            gene_seq = sequence[start - (upstream_flank_number + 1):end + downstream_flank_number]
            # Reverse complement this
            gene_seq = reverse_complement(gene_seq)
            upstream_flank = gene_seq[:upstream_flank_number]
            downstream_flank = gene_seq[-downstream_flank_number:]
            gene_seq = gene_seq[upstream_flank_number: -downstream_flank_number]

    return seqid, start, end, strand, upstream_flank, downstream_flank, gene_seq


def optimize_primer(plasmid_sequence, gene_sequence, desired_tm, direction, min_length=10, max_length=60,
                    tm_tolerance=5, max_temp_deviation=20, his_tag=False):
    """Optimize a primer sequence to achieve a desired Tm by adjusting its length."""

    best_primer = 'None'
    temp = 0
    if direction == 'forward':
        gene_sequence = gene_sequence
    elif direction == 'reverse':
        # Reverse comp it
        gene_sequence = str(Seq(gene_sequence).reverse_complement())
    else:
        u.dp([f'Direction: {direction} is not an option, must be: forward or reverse.'])
    if his_tag and direction == 'reverse':
        min_length = min_length + 18  # 33  # i.e. the 18 of his tag + the seq
    for primer_length in range(0, min_length, 3):
        for length in range(min_length, max_length, 3):
            plas_seq = plasmid_sequence[-1*(primer_length + 9):]
            temp_primer = plas_seq + gene_sequence[:length]
            tm = primer3.bindings.calcTm(temp_primer)
            temp_deviation = abs(desired_tm - tm)
            if temp_deviation < max_temp_deviation and temp_deviation < tm_tolerance and (temp_primer[-1] == 'C' or temp_primer[-1] == 'G'):
                max_temp_deviation = temp_deviation
                best_primer = plas_seq.lower() + gene_sequence[:length]
                temp = tm
    return best_primer, temp


def make_primer(gene, max_length=60, min_length=15, tm_tolerance=30, desired_tm=62.0,
                forward_primer='gaaataattttgtttaactttaagaaggagatatacat', reverse_primer='ctttgttagcagccggatc'):
    # Standard pET-22b(+) primer sequences
    forward_plasmid_primer = forward_primer.upper()  # Cut off a bit and then do the same for the tail
    # Play with the length to get the melting but you can optimise for around 38 length --> Full GCTTTGTTAGCAG  --> CCGGATCTCA
    reverse_plasmid_primer = reverse_primer.upper()
    # Desired Tm range for optimization
    # Target melting temperature in °C
    # Generate and optimize forward primer
    forward_gene_primer, forward_tm = optimize_primer(forward_plasmid_primer, gene, desired_tm, 'forward',
                                                      min_length, max_length, tm_tolerance)
    reverse_gene_primer, reverse_tm = optimize_primer(reverse_plasmid_primer, gene, desired_tm, 'reverse',
                                                      min_length, max_length, tm_tolerance, his_tag=True)

    print(f"Forward Gene Primer: 5'-{forward_gene_primer}-3' (Tm: {forward_tm:.2f} °C)")
    print(f"Reverse Gene Primer: 3'-{reverse_gene_primer}-5' (Tm: {reverse_tm:.2f} °C)")
    return forward_gene_primer, forward_tm, reverse_gene_primer, reverse_tm


def make_primers_IDT(fasta_file, remove_stop_codon=True, his_tag='',
                     max_length=60, min_length=15, tm_tolerance=30, desired_tm=62.0,
                     forward_primer='gaaataattttgtttaactttaagaaggagatatacat',
                     reverse_primer='ctttgttagcagccggatc'):
    """
    optionally would set the his_tag to be: 'CACCACCACCACCACCAC'
    # Note expects the stop codon to still be there since this gets removed and then the his tag added
    """
    seqs = {}
    for record in SeqIO.parse(fasta_file, "fasta"):
        try:
            seq_id = str(record.id)
            seqs[seq_id] = str(record.seq)
        except:
            print(record.description)
    rows = []
    for gene in seqs:
        gene_seq = seqs.get(gene)
        if remove_stop_codon:
            gene_seq = gene_seq[: -3]
        if his_tag:
            gene_seq = gene_seq + his_tag
        if gene_seq:
            forward_gene_primer, forward_tm, reverse_gene_primer, reverse_tm = make_primer(gene_seq,
                                                                                           max_length=max_length,
                                                                                           min_length=min_length,
                                                                                           tm_tolerance=tm_tolerance,
                                                                                           desired_tm=desired_tm,
                                                                                           forward_primer=forward_primer,
                                                                                           reverse_primer=reverse_primer)
            # Cut off the stop codon and add in the his Tag
            rows.append([f'5F_{gene}', forward_gene_primer, '25nm', 'STD'])
            rows.append([f'3R_{gene}', reverse_gene_primer, '25nm', 'STD'])
            print(len(forward_gene_primer), len(reverse_gene_primer))
        else:
            print(gene)
    primer_df = pd.DataFrame(rows)
    primer_df.columns = ['Name', 'Sequence', 'Scale', 'Purification']
    return primer_df

def generate_random_primer(length):
    letters = ['A', 'T', 'G', 'C']  # Includes both lowercase and uppercase letters
    result = ''.join(random.choice(letters) for i in range(length))
    return result
def get_high_tm_primers(length_min, length_max, optimal_tm, number_primers, num_samples=1000):
    rows = []
    for i in tqdm(range(length_min, length_max)):
        # Create num_samples primers of this length
        for j in range(num_samples):
            primer = generate_random_primer(i)
            primer_tm = primer3.bindings.calcTm(primer)
            rows.append([primer, primer_tm, abs(optimal_tm - primer_tm), i])
    df = pd.DataFrame(rows, columns=['Primer', 'Tm', 'Tm_diff', 'Length'])
    return df.sort_values(by=['Tm_diff']).head(number_primers)


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

"""
Author: Ariane Mora
Date: September 2024
"""
import math
import os
import pysam
from collections import defaultdict
from oligopoolio.primers import *


u = SciUtil()


def make_oligo_single(codon_optimized_fasta, forward_primer='gaaataattttgtttaactttaagaaggagatatacat', 
                      forward_primer_len=15, reverse_primer='gatccggctgctaacaaag', reverse_primer_len=15, max_len=320):
    records = list(SeqIO.parse(codon_optimized_fasta, "fasta"))
    rows = []
    for r in records:
        seq = str(r.seq)
        if seq[:3] == 'ATG' and len(seq) < max_len:
            rows.append([str(r.id), seq])
        else:
            u.dp([r.id, f'Issue: either did not have a methionine start or was longer than {max_len}.'
                        f' This has been omitted'])
                    
    # Now we want to drop any that are > 320 lengths 
    single_df = pd.DataFrame(rows)
    single_df.columns = ['id', 'seq']
    single_df['forward_primer'] = forward_primer[-1*forward_primer_len:]
    single_df['reverse_primer'] = reverse_primer[:reverse_primer_len]
    single_df['oligo'] = [f'{f}{seq}{r}' for f, seq, r in single_df[['forward_primer', 'seq', 'reverse_primer']].values]
    return single_df


def make_oligo_double(codon_optimized_fasta, forward_primer='gaaataattttgtttaactttaagaaggagatatacat', 
                      forward_primer_len=15, reverse_primer='gatccggctgctaacaaag', reverse_primer_len=15, max_len=640, 
                      overlap_len=9):

    records = list(SeqIO.parse(codon_optimized_fasta, "fasta"))
    rows = []
    fwd = forward_primer[-1*forward_primer_len:]
    bwd = reverse_primer[:reverse_primer_len]
    for r in records:
        seq_id = r.id
        seq = str(r.seq)
        if seq[:3] == 'ATG' and len(seq) < max_len:
            cut = len(seq)//2
            first_half = seq[:cut-overlap_len]
            if len(seq) % 2 == 0:
                second_half = seq[-cut+overlap_len:]
            else:
                second_half = seq[-cut + overlap_len - 1:]
            overlap = seq[cut-overlap_len:cut+overlap_len].lower()
            # Print out the tm
            tm_1 = primer3.bindings.calcTm(first_half)
            tm_2 = primer3.bindings.calcTm(second_half)
            tm_overlap = primer3.bindings.calcTm(overlap)
            u.dp(['TM of overlap:', tm_overlap])
            # Also reverse complement the overlap
            rev_comp_overlap = str(Seq(overlap).reverse_complement())
            rows.append([seq_id + '_1', seq, 'first', overlap, rev_comp_overlap, f'{fwd}{first_half}{overlap}', tm_1,
                         tm_overlap, f'{first_half}{overlap}{second_half}'])
            rows.append([seq_id + '_2', seq, 'second', overlap, rev_comp_overlap,
                         str(Seq(f'{overlap}{second_half}{bwd}').reverse_complement()), tm_2,
                         tm_overlap, f'{first_half}{overlap}{second_half}'])
        else:
            u.dp([seq_id, f'Issue: either did not have a methionine start or was longer than {max_len}. '
                          f'This has been omitted'])
                    
    # Now we want to drop any that are > 320 lengths 
    double_df = pd.DataFrame(rows)
    double_df.columns = ['id', 'seq', 'label', 'overlap', 'rev_comp_overlap', 'oligo', 'tm_part', 'tm_overlap', 'insert']
    return double_df


def make_splitty_oligo(codon_optimized_fasta, forward_primer='gaaataattttgtttaactttaagaaggagatatacat',
                       forward_primer_len=15, reverse_primer='gatccggctgctaacaaag',
                       reverse_primer_len=15, max_len=640, max_part_len=100, overlap_len=18):
    records = list(SeqIO.parse(codon_optimized_fasta, "fasta"))
    rows = []
    fwd = forward_primer[-1 * forward_primer_len - 1:]
    bwd = reverse_primer[:reverse_primer_len]
    for r in records:
        max_part_len = 100
        seq_id = r.id
        seq = fwd + str(r.seq)
        if seq[-3:] == 'TAA':
            print('Stop codon removed...')
            seq = seq[:-3]
            print(seq[-10:])
        # Basically get the splits size
        total_seq = fwd + seq + bwd
        # get the number of splits
        num_splits = int(math.ceil(len(total_seq) / max_part_len))
        # Make sure it is even...
        if num_splits % 2 != 0:
            num_splits += 1
        reverse = True
        prev_overlap = ''
        # Make sure our cuts are
        max_part_len = int(len(total_seq)/num_splits)
        print(max_part_len, num_splits, len(total_seq))
        for i in range(0, num_splits):
            cut = max_part_len * (i + 1)
            prev_cut = max_part_len * i
            if i + 1 == num_splits:
                oligo = seq[prev_cut:]
                # Add the backwards last bit on
                if not reverse:
                    # Forward
                    rows.append([seq_id + '_end', 'forward_end', f'{oligo}{bwd}', prev_cut, cut, prev_overlap, bwd, oligo, len(f'{oligo}{bwd}')])
                    reverse = True
                    break
                elif reverse:
                    # reverse
                    rev_comp = str(Seq(f'{oligo}{bwd}').reverse_complement())
                    rows.append([seq_id + '_end', 'reverse_end', rev_comp, prev_cut, cut, prev_overlap, bwd, oligo, len(f'{oligo}{bwd}')])
                    reverse = False
                    break
            else:
                oligo = seq[prev_cut + 1:cut]
                next_overlap = seq[cut:cut + overlap_len]
                # Also reverse complement the overlap
                if not reverse:
                    # Forward
                    rows.append([seq_id + '_' + str(i), 'forward', f'{oligo}{next_overlap}', prev_cut, cut, prev_overlap, next_overlap, oligo, len(f'{oligo}{next_overlap}')])
                    reverse = True
                elif reverse:
                    # reverse
                    rev_comp = str(Seq(f'{oligo}{next_overlap}').reverse_complement())
                    rows.append([seq_id + '_' + str(i), 'reverse', rev_comp, prev_cut, cut, prev_overlap, next_overlap, oligo, len(rev_comp)])
                    reverse = False
                prev_overlap = next_overlap

        # Now we want to drop any that are > 320 lengths
    double_df = pd.DataFrame(rows)
    double_df.columns = ['id', 'label', 'oligo', 'prev_cut', 'cut', 'prev_overlap', 'next_overlap', 'oligo', 'length']
    return double_df


# Take the demultiplexed files from a LevSeq run and then demultiplex oligoPools from these

def alignment_from_cigar(cigar: str, alignment: str, ref: str, query_qualities: list):
    """
    Generate the alignment from the cigar string.
    Operation	Description	Consumes query	Consumes reference
    0 M	alignment match (can be a sequence match or mismatch)	yes	yes
    1 I	insertion to the reference	yes	no
    2 D	deletion from the reference	no	yes
    3 N	skipped region from the reference	no	yes
    4 S	soft clipping (clipped sequences present in SEQ)	yes	no
    5 H	hard clipping (clipped sequences NOT present in SEQ)	no	no
    6 P	padding (silent deletion from padded reference)	no	no
    7 =	sequence match	yes	yes
    8 X	sequence mismatch	yes	yes
    """
    new_seq = ''
    ref_seq = ''
    qual = []
    inserts = []
    pos = 0
    ref_pos = 0
    for op, op_len in cigar:
        if op == 0:  # alignment match (can be a sequence match or mismatch)
            new_seq += alignment[pos:pos + op_len]
            qual += query_qualities[pos:pos + op_len]

            ref_seq += ref[ref_pos:ref_pos + op_len]
            pos += op_len
            ref_pos += op_len
        elif op == 1:  # insertion to the reference
            inserts.append(alignment[pos - 1:pos + op_len])
            pos += op_len
        elif op == 2:  # deletion from the reference
            new_seq += '-' * op_len
            qual += [-1] * op_len
            ref_seq += ref[ref_pos:ref_pos + op_len]
            ref_pos += op_len
        elif op == 3:  # skipped region from the reference
            new_seq += '*' * op_len
            qual += [-2] * op_len
            ref_pos += op_len
        elif op == 4:  # soft clipping (clipped sequences present in SEQ)
            inserts.append(alignment[pos:pos + op_len])
            pos += op_len
        elif op == 5:  # hard clipping (clipped sequences NOT present in SEQ)
            continue
        elif op == 6:  # padding (silent deletion from padded reference)
            continue
        elif op == 7:  # sequence mismatch
            new_seq += alignment[pos:pos + op_len]
            ref_seq += ref[ref_pos:ref_pos + op_len]
            qual += query_qualities[pos:pos + op_len]
            pos += op_len
            ref_pos += op_len
    return new_seq, ref_seq, qual, inserts


def annotate_to_wells(plate_path, ref_fasta):
    """ 
    This is for annotating the demuiltiplexed files from LevSeq to the wells that it came from 
    Assuming that you've demultiplexed the long read reads from a 96 well plate using the levSeq formatting.
    """
    files = os.listdir(plate_path)
    # Then just go through each well and annotate these using minimap2
    # make a mapping between the read id and the sequence so that this can be used as the reference string when aligning
    ref_map = {}
    seqs = [str(record.seq) for record in SeqIO.parse(ref_fasta, "fasta")]
    ref_read_ids = [str(record.id) for record in SeqIO.parse(ref_fasta, "fasta")]
    for i, seq in enumerate(seqs):
        ref_map[ref_read_ids[i]] = seq
    rows_all = []
    for well in files:
        try:
            msa_path = f'{plate_path}/{well}/{well}.msa'
            # get the file from there
            os.system(f'minimap2 -ax map-ont -A 4 -B 2 -O 10,24 {ref_fasta} {plate_path}/{well}/demultiplexed_RB07_{well}_000.fastq > {plate_path}/{well}/{well}.sam')
            os.system(f'samtools view -bS {plate_path}/{well}/{well}.sam > {plate_path}/{well}/{well}.bam')
            os.system(f'samtools sort {plate_path}/{well}/{well}.bam -o {plate_path}/{well}/{well}.bam')
            os.system(f'samtools index {plate_path}/{well}/{well}.bam')
            bam_file_path = f'{plate_path}/{well}/{well}.bam'
            bam = pysam.AlignmentFile(bam_file_path, "rb")
            # Ensure the BAM file is indexed
            if not os.path.exists(bam_file_path + ".bai"):
                pysam.index(bam_file_path)
            
            seqs = []
            read_ids = []
            read_quals = []
            err = 0
            read_counts = defaultdict(int)

            for read in bam.fetch(until_eof=True):
                try:
                    ref_str = ref_map.get(read.reference_name)
                    if read.query_sequence is not None and len(read.query_sequence) > 0.9*len(ref_str) and read.cigartuples is not None:
                        seq, ref, qual, ins = alignment_from_cigar(read.cigartuples, read.query_sequence, ref_str,
                                                                read.query_qualities)
                        # Make it totally align
                        seq = "-" * read.reference_start + seq + "-" * (len(ref_str) - (read.reference_start + len(seq)))
                        seqs.append(seq)
                        #seqs.append(read.query_sequence)
                        read_ids.append(f'{read.reference_name}')
                        read_quals.append(read.qual)
                        # keep track of how many reads there were
                        read_counts[read.reference_name] += 1
                except:
                    err += 1
            print(err)
            # Check if we want to write a MSA
            if msa_path is not None:
                print("Writing MSA")
                with open(msa_path, 'w+') as fout:
                    # Write the reference first
                    fout.write(f'>{read.query_name}\n{ref_str}\n')
                    for i, seq in enumerate(seqs):
                        fout.write(f'>{read_ids[i]}\n{"".join(seq)}\n')
                # Align using clustal for debugging if you need the adapter! Here you would change above to use a different version
                print(f'clustal-omega --force -i {msa_path} -o {msa_path.replace(".fa", "_msa.fa")}')
                os.system(f'clustal-omega --force -i "{msa_path}" -o "{msa_path.replace(".fa", "_msa.fa")}"')
                seqs = [str(record.seq) for record in SeqIO.parse(msa_path.replace(".fa", "_msa.fa"), "fasta")]
                read_ids = [str(record.id) for record in SeqIO.parse(msa_path.replace(".fa", "_msa.fa"), "fasta")]
        except:
            print(well)
        # Again check that we actually had enough reads for this to be considered a good well
        # Count how many unique read IDs mapped to this (we hope that it is all the same )
        bam.close()
        # Get the number for each read in there
        row = []
        most_read = 0
        well_id = None
        for read in ref_read_ids:
            row.append(read_counts.get(read))
            if read_counts.get(read) and read_counts.get(read) > most_read:
                most_read = read_counts.get(read)
                well_id = read
        rows_all.append([well, len(read_ids), well_id, most_read] + row) #, np.mean(read_quals)])

    if len(rows_all) > 1:  # Check if we have anything to return
        seq_df = pd.DataFrame(rows_all, columns=['Well', 'Number of reads', 'Assigned ID', 'Reads for Assigned ID'] + ref_read_ids)
        seq_df.to_csv('cutadapt_demultiplex_seqs.csv', index=False)




from setuptools import setup
import os
import re


def read_version():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'oligopoolio/__init__.py')
    with open(path, 'r') as fh:
        return re.search(r'__version__\s?=\s?[\'"](.+)[\'"]', fh.read()).group(1)


def read_author():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'oligopoolio/__init__.py')
    with open(path, 'r') as fh:
        return re.search(r'__author__\s?=\s?[\'"](.+)[\'"]', fh.read()).group(1)


def read_email():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'oligopoolio/__init__.py')
    with open(path, 'r') as fh:
        return re.search(r'__author_email__\s?=\s?[\'"](.+)[\'"]', fh.read()).group(1)


def read_git():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'oligopoolio/__init__.py')
    with open(path, 'r') as fh:
        return re.search(r'__url__\s?=\s?[\'"](.+)[\'"]', fh.read()).group(1)


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='oligopoolio',
      version=read_version(),
      description='',
      long_description=readme(),
      long_description_content_type='text/markdown',
      author=read_author(),
      author_email=read_email(),
      url=read_git(),
      license='GPL3',
      project_urls={
          "Bug Tracker": read_git(),
          "Documentation": read_git(),
          "Source Code": read_git()},
      classifiers=[
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords=['oligopools', 'protein-engineering'],
      packages=['oligopoolio'],
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'oligopoolio = oligopoolio.__main__:main'
          ]
      },
      install_requires=['scikit-learn',
                        'dnaweaver',
                        'dnachisel',
                        'pdf_reports',
                        'sciutil',
                        'seaborn',
                        'statsmodels',
                        'biopython',
                        'primer3-py',
                        'pandas',
                        'numpy',
                        'gffutils',
                        'jupyterlab',
                        'typer'],
      python_requires='>=3.8',
      data_files=[("", ["LICENSE"])]
      )



import dnaweaver as dw
import time
import dnachisel as dnachisel
from sciutil import SciUtil
from Bio.Seq import Seq
from difflib import SequenceMatcher
import primer3
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
import pandas as pd
from primer3 import calc_hairpin, calc_homodimer
import math
from dnachisel import *
import pyswarms as ps
import Levenshtein
import numpy as np
from functools import partial
import seaborn as sns
import matplotlib.pyplot as plt
from jinja2 import Template
from weasyprint import HTML
import io
import base64
import matplotlib
import os
from dna_features_viewer import BiopythonTranslator
from Bio import SeqIO
import matplotlib.pyplot as plt
from tqdm import tqdm 

u = SciUtil()

# Use Arial
matplotlib.rcParams["font.family"] = "Arial"
sns.set(style="ticks")


def plot_to_base64(fig):
    buf = io.BytesIO()
    fig.savefig(buf, format="png", bbox_inches="tight")
    buf.seek(0)
    return base64.b64encode(buf.read()).decode("utf-8")


def plot_linear_section_from_gb(
    file_list, seq_list, feature_start=1, feature_filter=None
):
    """
    Returns base64-encoded plots of GenBank features in linear section.
    """
    plots = []

    class CustomTranslator(BiopythonTranslator):
        def compute_feature_color(self, feature):
            return "#ffd700" if "primer" in feature.type.lower() else "#87cefa"

        def compute_filtered_features(self, features):
            if feature_filter:
                return [f for f in features if feature_filter(f)]
            return features

    for i, file_path in enumerate(file_list):
        record = SeqIO.read(file_path, "genbank")
        translator = CustomTranslator()
        graphic_record = translator.translate_record(record)

        graphic_record = graphic_record.crop(
            (feature_start - 20, feature_start + len(seq_list[i]) + 50)
        )

        fig, ax = plt.subplots(figsize=(10, 2))
        graphic_record.plot(ax=ax)
        ax.set_title(f"{os.path.basename(file_path)}")
        img_base64 = plot_to_base64(fig)
        plt.close(fig)

        plots.append({"name": os.path.basename(file_path), "image": img_base64})

    return plots


def generate_pdf_report(
    oligo_df,
    seq_list,
    gb_file_list,
    insert_position,
    min_similarity=8,
    output_pdf_path="oligo_report.pdf",
):
    """Make a PDF since it's easier than looking through..."""
    overlaps = [
        x for x in oligo_df["primer_overlap_with_previous"].values if x is not None
    ]
    ids = oligo_df["id"].unique()
    plt.rcParams["svg.fonttype"] = "none"  # Ensure text is saved as text
    plt.rcParams["figure.figsize"] = (3, 3)
    sns.set(
        rc={
            "figure.figsize": (3, 3),
            "font.family": "sans-serif",
            "font.sans-serif": "Arial",
            "font.size": 12,
        },
        style="ticks",
    )
    # Summary stats
    summaries = []
    for seq_id, grp in oligo_df.groupby("id"):
        summaries.append(
            {
                "id": seq_id,
                "min_tm": np.nanmin(grp["overlap_tm_5prime"]),
                "max_tm": np.nanmax(grp["overlap_tm_5prime"]),
                "min_dg": np.nanmin(grp["overlap_homodimer_dg"]),
                "min_len": np.nanmin(grp["oligo_length"]),
                "max_len": np.nanmax(grp["oligo_length"]),
                "min_ovl": np.nanmin(grp["overlap_length"]),
                "max_ovl": np.nanmax(grp["overlap_length"]),
                "num_fragments": len(grp["overlap_length"].values),
                "best_cost": np.nanmin(grp["best_cost"]),  # Always the same...
            }
        )

    # Set seaborn style globally
    sns.set(style="ticks", context="notebook", font="Arial")
    # Extract overlaps and their corresponding IDs
    overlaps = [
        x for x in oligo_df["primer_overlap_with_previous"].values if x is not None
    ]
    overlap_ids = oligo_df[~oligo_df["primer_overlap_with_previous"].isna()][
        "oligo_id"
    ].tolist()
    n = len(overlaps)

    # Build full distance matrix
    dist_matrix = np.full((n, n), np.nan)
    for i in range(n):
        for j in range(n):
            if i != j:
                dist = Levenshtein.distance(overlaps[i], overlaps[j])
                if dist < min_similarity:
                    dist_matrix[i, j] = dist

    # Create labeled DataFrame
    dist_df = pd.DataFrame(dist_matrix, index=overlap_ids, columns=overlap_ids)

    # Drop rows/cols where all values are NaN
    dist_df = dist_df.dropna(axis=0, how="all").dropna(axis=1, how="all")

    # Plot only if there's something to show
    if not dist_df.empty:
        
        fig1, ax1 = plt.subplots(
            figsize=(max(6, dist_df.shape[0]), max(5, dist_df.shape[1]))
        )
        sns.heatmap(
            dist_df,
            annot=True,
            fmt=".0f",
            cmap="coolwarm",
            cbar=True,
            square=True,
            ax=ax1,
            linewidths=0.5,
            linecolor="lightgrey",
        )
        ax1.set_title(f"Levenshtein Distances < {min_similarity}", fontsize=14)
        ax1.set_xlabel("Sequence ID")
        ax1.set_ylabel("Sequence ID")
        plt.xticks(rotation=45, ha="right")
        plt.yticks(rotation=0)
        ax1 = clean_plt(ax1)
        heatmap_base64 = plot_to_base64(fig1)
        plt.savefig('heatmap.png')
        plt.close(fig1)
        
    else:
        heatmap_base64 = None  # You can handle this later in the HTML/PDF
    # --- Temperature Plot ---
    fig2, ax2 = plt.subplots(figsize=(8, 4))
    df_temp = pd.DataFrame(summaries)
    sns.lineplot(
        x="id",
        y="min_tm",
        data=df_temp,
        marker="o",
        label="Min Tm",
        ax=ax2,
        color="blue",
    )
    sns.lineplot(
        x="id",
        y="max_tm",
        data=df_temp,
        marker="o",
        label="Max Tm",
        ax=ax2,
        color="lightblue",
    )
    ax2.set_title("Overlap Tm per ID", fontsize=14)
    ax2.set_ylabel("Temperature (°C)")
    ax2.set_xlabel("ID")
    ax2 = clean_plt(ax2)
    temp_plot_base64 = plot_to_base64(fig2)
    plt.close(fig2)

    # --- Length Plot ---
    fig3, ax3 = plt.subplots(figsize=(8, 4))
    sns.lineplot(
        x="id",
        y="min_len",
        data=df_temp,
        marker="s",
        label="Min Length",
        ax=ax3,
        color="green",
    )
    sns.lineplot(
        x="id",
        y="max_len",
        data=df_temp,
        marker="s",
        label="Max Length",
        ax=ax3,
        color="lightgreen",
    )
    ax3.set_title("Oligo Length per ID", fontsize=14)
    ax3.set_ylabel("Length (bp)")
    ax3.set_xlabel("ID")
    ax3 = clean_plt(ax3)
    len_plot_base64 = plot_to_base64(fig3)
    plt.close(fig3)

    gb_section_plots = plot_linear_section_from_gb(
        file_list=gb_file_list,
        seq_list=seq_list,
        feature_start=insert_position,  # Optional
        # feature_filter=lambda f: "seq" in f.type.lower()  # Optional
    )
    # HTML with inline CSS for Arial
    html_template = """
    <html>
    <head>
        <style>
            body { font-family: Arial, sans-serif; padding: 20px; }
            h1, h2 {font-family: Arial, sans-serif; color: #333; }
            table { font-family: Arial, sans-serif; border-collapse: collapse; width: 100%; margin-bottom: 20px; }
            th, td { border: 1px solid #aaa; padding: 8px; text-align: center; }
            th { background-color: #f2f2f2; }
            img { max-width: 100%; height: auto; margin-bottom: 20px; }
        </style>
    </head>
    <body>
        <h1>Oligo Evaluation Report</h1>

        <h2>Summary Table</h2>
        <table>
            <tr>
                <th>ID</th><th>Min Tm</th><th>Max Tm</th><th>Min ΔG</th>
                <th>Min Length (w. primer)</th><th>Max Length (w. primer)</th><th>Min Overlap</th><th>Max Overlap</th><th>Best cost</th><th># Fragments</th>
            </tr>
            {% for s in summaries %}
            <tr>
                <td>{{ s.id }}</td>
                <td>{{ "%.2f"|format(s.min_tm) }}</td>
                <td>{{ "%.2f"|format(s.max_tm) }}</td>
                <td>{{ "%.2f"|format(s.min_dg) }}</td>
                <td>{{ "%.0f"|format(s.min_len) }}</td>
                <td>{{ "%.0f"|format(s.max_len) }}</td>
                <td>{{ "%.0f"|format(s.min_ovl) }}</td>
                <td>{{ "%.0f"|format(s.max_ovl) }}</td>
                <td>{{ "%.0f"|format(s.best_cost) }}</td>
                <td>{{ "%.0f"|format(s.num_fragments) }}</td>
            </tr>
            {% endfor %}
        </table>

        <h2>Temperature Plot</h2>
        <img src="data:image/png;base64,{{ temp_plot_base64 }}" />

        <h2>Oligo Length Plot</h2>
        <img src="data:image/png;base64,{{ len_plot_base64 }}" />

        <h2>Overlap Similarity Heatmap</h2>

        <h2>GenBank Linear Section Plots</h2>
        {% for gb in gb_plots %}
          <h3>{{ gb.name }}</h3>
          <img src="data:image/png;base64,{{ gb.image }}" />
        {% endfor %}

    </body>
    </html>
    """

    template = Template(html_template)
    html_content = template.render(
        summaries=summaries,
        heatmap_base64=heatmap_base64,
        temp_plot_base64=temp_plot_base64,
        len_plot_base64=len_plot_base64,
        min_similarity=min_similarity,
        gb_plots=gb_section_plots,
    )

    # Save to PDF
    HTML(string=html_content).write_pdf(output_pdf_path)
    print(f"PDF report saved as: {output_pdf_path}")


def clean_plt(ax):
    ax.tick_params(direction="out", length=2, width=1.0)
    ax.spines["bottom"].set_linewidth(1.0)
    ax.spines["top"].set_linewidth(0)
    ax.spines["left"].set_linewidth(1.0)
    ax.spines["right"].set_linewidth(0)
    ax.tick_params(labelsize=10.0)
    ax.tick_params(axis="x", which="major", pad=2.0)
    ax.tick_params(axis="y", which="major", pad=2.0)
    return ax


def create_oligos(
    seq_df,
    seq_col,
    id_col,
    genbank_file,
    bsa,
    min_overlap,
    max_overlap,
    optimal_seq_len,
    min_seq_len,
    max_seq_len,
    tm_tolerance,
    reverse=False,
    insert_position=5193,
    backbone_5_overlap="",
    backbone_3_overlap="",
    sequence_end="",
    num_rounds=10
):
    rows = []
    for seq, seq_id in tqdm(seq_df[[seq_col, id_col]].values):
        # Add the end on
        seq = seq + sequence_end
        # Optimize fragments with the backbone overlap
        oligos, best_cost = optimize_fragments_for_gene(
            backbone_5_overlap + seq + backbone_3_overlap,
            min_overlap,
            max_overlap,
            optimal_seq_len,
            min_seq_len,
            max_seq_len,
            tm_tolerance,
            num_rounds
        )
        translation_label = f"Insert_{seq_id}"
        record = insert_sequence_with_translation(
            genbank_file, None, insert_position, seq, translation_label, reverse
        )
        prev_oligo = None
        if len(oligos) > 0:
            for i, oligo in enumerate(oligos):
                seq = oligo
                primer_overlap, primer_tm, primer_len, homodimer_tm, hairpin_tm = (
                    None,
                    None,
                    None,
                    None,
                    None,
                )
                if prev_oligo:
                    match = SequenceMatcher(None, prev_oligo, seq).find_longest_match()
                    primer_overlap = prev_oligo[match.a : match.a + match.size]
                    results = check_secondary_structure(primer_overlap)
                    homodimer_tm = results["homodimer"]["homodimer_dg"]
                    hairpin_tm = results["hairpin"]["hairpin_dg"]
                    primer_tm = primer3.bindings.calcTm(primer_overlap)
                    primer_len = len(primer_overlap)
                prev_oligo = seq
                orig_seq = seq
                strand = 1
                if bsa:
                    if i % 2 == 1 and i != 0:
                        seq = str(Seq(seq).reverse_complement())
                        strand = -1
                else:
                    if i % 2 == 0:
                        seq = str(Seq(seq).reverse_complement())
                        strand = -1
                oligo_tm = primer3.bindings.calcTm(seq)
                insert_features_from_oligos(
                    record, f"{seq_id}_oligo_{i}", orig_seq, strand, oligo_tm, None
                )
                rows.append(
                    [
                        seq_id,
                        f"{seq_id}_oligo_{i}",
                        seq,
                        len(seq),
                        oligo_tm,
                        strand,
                        primer_overlap,
                        primer_tm,
                        primer_len,
                        homodimer_tm,
                        hairpin_tm,
                        seq,
                        best_cost,
                    ]
                )
            if genbank_file:
                output_file = genbank_file.replace(".", f"_{seq_id}.")
                record.name = seq_id
                SeqIO.write(record, f"{output_file}", "genbank")
    oligo_df = pd.DataFrame(
        rows,
        columns=[
            "id",
            "oligo_id",
            "oligo_sequence",
            "oligo_length",
            "oligo_tm",
            "strand",
            "primer_overlap_with_previous",
            "overlap_tm_5prime",
            "overlap_length",
            "overlap_homodimer_dg",
            "overlap_hairpin_dg",
            "original_sequence",
            "best_cost",
        ],
    )
    return oligo_df


def optimize_fragments_for_gene(
    seq,
    min_overlap,
    max_overlap,
    optimal_seq_len,
    min_seq_len,
    max_seq_len,
    tm_tolerance,
    num_rounds
):
    seq_len = len(seq)
    num_fragments = int(math.ceil(seq_len / optimal_seq_len))
    if num_fragments % 2 == 1:
        num_fragments += 1
    num_cuts = num_fragments - 1
    num_variables = num_cuts * 2

    cut_lower_bounds = np.arange(1, num_cuts + 1) * 2
    cut_upper_bounds = np.full(num_cuts, seq_len - 2)
    olap_lower_bounds = np.full(num_cuts, min_overlap)
    olap_upper_bounds = np.full(num_cuts, max_overlap)

    lower_bounds = np.concatenate([cut_lower_bounds, olap_lower_bounds])
    upper_bounds = np.concatenate([cut_upper_bounds, olap_upper_bounds])

    bounds = (lower_bounds, upper_bounds)

    total_best_cost = 100000
    total_best_pos = None
    # Do a few runs so that we converge sometimes it gets stuck
    for run in range(0, num_rounds):
        optimizer = ps.single.GlobalBestPSO(
            n_particles=150,
            dimensions=num_variables,
            options={"c1": 0.5, "c2": 0.5, "w": 0.7},
            bounds=bounds,
        )
        best_cost, best_pos = optimizer.optimize(
            partial(
                objective_function,
                seq_len=seq_len,
                num_fragments=num_fragments,
                seq=seq,
                max_overlap=max_overlap,
                min_overlap=min_overlap,
                max_seq_len=max_seq_len,
                min_seq_len=min_seq_len,
                tm_tolerance=tm_tolerance,
            ),
            iters=100,
        )
        if best_cost < total_best_cost:
            total_best_cost = best_cost
            total_best_pos = best_pos

    best_cost = total_best_cost
    best_pos = total_best_pos

    cuts = np.sort(np.round(best_pos[:num_cuts]).astype(int))  # type: ignore
    overlaps = np.round(best_pos[num_cuts:]).astype(int)  # type: ignore
    fragments = get_fragments(cuts, seq_len, num_fragments, overlaps)
    fragment_strings = [seq[start:end] for start, end in fragments]

    return fragment_strings, best_cost


def get_fragments(cuts, seq_len, num_fragments, overlap_lens):
    cuts = np.sort(np.round(cuts).astype(int))
    starts = [0] + list(cuts)
    fragments = []

    for i in range(num_fragments):
        start = starts[i]
        if i < num_fragments - 1:
            next_start = starts[i + 1]
            end = next_start + overlap_lens[i]
        else:
            end = seq_len
        end = min(end, seq_len)
        fragments.append((int(start), int(end)))
    return fragments


def objective_function(
    x,
    seq_len,
    num_fragments,
    seq,
    tm_optimal=62,
    min_dist=8,
    max_overlap=27,
    min_overlap=18,
    min_seq_len=80,
    max_seq_len=130,
    tm_tolerance=5,
):
    penalties = []
    for particle in x:
        cuts = particle[: num_fragments - 1]
        overlaps = particle[num_fragments - 1 :]
        fragments = get_fragments(cuts, seq_len, num_fragments, overlaps)
        frag_lengths = [end - start for start, end in fragments]

        olap_penalty = 0
        overlaps_seqs = []
        temp_penalty = 0
        seq_len_penalty = 0
        homodimer_penalty = 0
        repeat_penalty = 0
        for i in range(len(fragments) - 1):
            start1, end1 = fragments[i]
            start2, end2 = fragments[i + 1]
            overlap_seq = seq[start2:end1]
            results = check_secondary_structure(overlap_seq)
            homodimer_tm = results["homodimer"]["homodimer_dg"]
            hairpin_tm = results["hairpin"]["hairpin_dg"]
            primer_tm = primer3.bindings.calcTm(overlap_seq)

            tm_penalty = np.abs(primer_tm - tm_optimal)
            if len(overlap_seq) < min_overlap or len(overlap_seq) > max_overlap:
                olap_penalty += 10 * abs(
                    len(overlap_seq) - (min_overlap + max_overlap) / 2
                )

            temp_penalty += 10 * tm_penalty if tm_penalty > tm_tolerance else tm_penalty

            for frag_len in [end1 - start1, end2 - start2]:
                if frag_len > max_seq_len:
                    seq_len_penalty += 10 * abs(max_seq_len - frag_len)
                elif frag_len < min_seq_len:
                    seq_len_penalty += 10 * abs(min_seq_len - frag_len)

            homodimer_penalty += (
                10 * abs(homodimer_tm) if homodimer_tm < -3 else -1 * homodimer_tm
            )
            if overlap_seq[:3] in ["AAA", "TTT", "CCC", "GGG"] or overlap_seq[-3:] in [
                "AAA",
                "TTT",
                "CCC",
                "GGG",
            ]:
                repeat_penalty += 20
            overlaps_seqs.append(overlap_seq)

        min_dist_val = 100000
        for i, x in enumerate(overlaps_seqs):
            for j, y in enumerate(overlaps_seqs):
                if i != j:
                    dist = Levenshtein.distance(x, y)
                    if dist < min_dist_val:
                        min_dist_val = dist
        dist_penalty = max_overlap - min_dist_val
        if min_dist_val < min_dist:
            dist_penalty += max_overlap

        mean_len = np.mean(frag_lengths)
        variance = np.sqrt(np.sum([(l - mean_len) ** 2 for l in frag_lengths]))

        penalties.append(
            variance
            + olap_penalty
            + homodimer_penalty
            + temp_penalty
            + seq_len_penalty
            + dist_penalty
            + repeat_penalty
        )
    return np.array(penalties)


def insert_sequence_with_translation(
    input_file,
    output_file,
    insert_position,
    new_sequence,
    translation_label,
    reverse=False,
):
    """
    Insert a new sequence at a specific position in a GenBank file, add a translation annotation,
    and adjust existing annotations to avoid overlap.

    Args:
        input_file (str): Path to the original GenBank file.
        output_file (str): Path to save the modified GenBank file.
        insert_position (int): Position to insert the new sequence (0-based).
        new_sequence (str): The DNA sequence to insert.
        translation_label (str): Label for the translation feature.
        reverse (bool): Whether the feature should be on the reverse strand.
    """
    # Read the original GenBank file
    record = SeqIO.read(input_file, "genbank")

    # Insert the new sequence at the specified position
    if reverse:
        new_sequence = str(
            Seq(new_sequence).reverse_complement()
        )  # Reverse complement the sequence if needed
    record.seq = (
        record.seq[:insert_position] + Seq(new_sequence) + record.seq[insert_position:]
    )

    # Adjust existing annotations to avoid overlap
    inserted_length = len(new_sequence)
    for feature in record.features:
        if feature.location.start >= insert_position:
            feature.location = FeatureLocation(
                feature.location.start + inserted_length,
                feature.location.end + inserted_length,
                strand=feature.location.strand,
            )

    # Create the feature label
    strand_label = " (reverse)" if reverse else " (forward)"
    full_label = translation_label + strand_label

    # Add a feature for the inserted sequence
    start = insert_position
    end = insert_position + len(new_sequence)
    feature = SeqFeature(
        location=FeatureLocation(start, end, strand=-1 if reverse else 1),
        type="CDS",  # CDS type for coding sequences
        qualifiers={
            "label": full_label,
            "translation": Seq(new_sequence).translate(
                table=11 # type: ignore
            ),  # Translation for the sequence
        },
    )
    record.features.append(feature)
    # Save the modified GenBank file
    if output_file:
        SeqIO.write(record, output_file, "genbank")
        print(f"Updated GenBank file saved as {output_file}")

    return record


def insert_features_from_oligos(record, seq_id, seq, strand, tm, output_file):
    """
    Insert features into a GenBank file based on oligo_df.

    Args:
        genbank_file (str): Path to the original GenBank file.
        output_file (str): Path to save the updated GenBank file.
        oligo_df (pd.DataFrame): DataFrame with oligo information.
                                 Expected columns: ['id', 'oligo_id', 'oligo_sequence', 'oligo_length', ...].
    """
    # Iterate through the oligo DataFrame to add features
    start = record.seq.find(seq.upper())
    if start == -1:
        print(f"Warning: Oligo sequence {seq_id} not found in the GenBank sequence.")

    end = start + len(seq)
    feature = SeqFeature(
        location=FeatureLocation(start, end, strand=strand),
        type="misc_feature",
        qualifiers={
            "label": f"{seq_id} {'(reverse)' if strand == -1 else '(forward)'}",
            "note": f"Length: {len(seq)}, TM: {tm}",
        },
    )
    record.features.append(feature)

    # Save the updated GenBank file
    if output_file:
        SeqIO.write(record, output_file, "genbank")
        print(f"Updated GenBank file saved as {output_file}")
    return record


def check_secondary_structure(sequence, temp=55):
    """
    Check secondary structures like hairpins and homodimers in a given primer sequence.

    Args:
        sequence (str): The DNA sequence of the primer to analyze.

    Returns:
        dict: Results for hairpin and homodimer properties.
    """
    try:
        # Check for hairpin structure
        hairpin_result = calc_hairpin(sequence, temp_c=temp)
        hairpin_info = {
            "hairpin_found": hairpin_result.structure_found,
            "hairpin_tm": hairpin_result.tm,
            "hairpin_dg": hairpin_result.dg / 1000,
            "hairpin_dh": hairpin_result.dh / 1000,
            "hairpin_ds": hairpin_result.ds,
        }

        # Check for homodimer structure
        homodimer_result = calc_homodimer(sequence, temp_c=temp)
        homodimer_info = {
            "homodimer_found": homodimer_result.structure_found,
            "homodimer_tm": homodimer_result.tm,
            "homodimer_dg": homodimer_result.dg / 1000,
            "homodimer_dh": homodimer_result.dh / 1000,
            "homodimer_ds": homodimer_result.ds,
        }
    except Exception as e:
        u.warn_p([f"Warning: issue with secondary structure check. ", sequence, e])
        hairpin_info = {
            "hairpin_found": False,
            "hairpin_tm": None,
            "hairpin_dg": None,
            "hairpin_dh": None,
            "hairpin_ds": None,
        }
        homodimer_info = {
            "homodimer_found": False,
            "homodimer_tm": None,
            "homodimer_dg": None,
            "homodimer_dh": None,
            "homodimer_ds": None,
        }
    # Combine results
    return {"hairpin": hairpin_info, "homodimer": homodimer_info}


def build_oligos_DNAWeaver(
    seq_id: str,
    sequence: str,
    output_directory: str,
    min_gc=0.3,
    max_gc=0.7,
    min_tm=55,
    max_tm=70,
    min_segment_length=40,
    max_segment_length=100,
    max_length=1500,
):
    """Use DNAweaver to build oligos"""
    # Here we use a comercial supplier but don't actually care.
    cheap_dna_offer = dw.CommercialDnaOffer(
        name="CheapDNA.",
        sequence_constraints=[
            dw.NoPatternConstraint(enzyme="BsaI"),
            dw.SequenceLengthConstraint(max_length=4000),
            dw.GcContentConstraint(min_gc=min_gc, max_gc=max_gc), # type: ignore
        ],
        pricing=dw.PerBasepairPricing(0.10),
    )

    oligo_dna_offer = dw.CommercialDnaOffer(
        name="OliGoo",
        sequence_constraints=[
            dw.GcContentConstraint(min_gc=min_gc, max_gc=max_gc), # type: ignore
            dw.SequenceLengthConstraint(max_length=4000),
        ],
        pricing=dw.PerBasepairPricing(0.07),
        memoize=True,
    )

    oligo_assembly_station = dw.DnaAssemblyStation(
        name="Oligo Assembly Station",
        assembly_method=dw.OligoAssemblyMethod(
            overhang_selector=dw.TmSegmentSelector(
                min_size=15, max_size=25, min_tm=min_tm, max_tm=max_tm
            ),
            min_segment_length=min_segment_length,
            max_segment_length=max_segment_length,
            sequence_constraints=[dw.SequenceLengthConstraint(max_length=4000)],
            duration=8,
            cost=30,
        ),
        supplier=oligo_dna_offer,
        coarse_grain=20,
        a_star_factor="auto",
        memoize=True,
    )

    assembly_station = dw.DnaAssemblyStation(
        name="Gibson Assembly Station",
        assembly_method=dw.GibsonAssemblyMethod(
            overhang_selector=dw.TmSegmentSelector(min_tm=min_tm, max_tm=max_tm),
            min_segment_length=min_segment_length,
            max_segment_length=max_segment_length + 20,  # add a bit of a buffer
        ),
        supplier=[cheap_dna_offer, oligo_assembly_station],
        logger="bar",
        coarse_grain=100,
        fine_grain=10,
        a_star_factor="auto",
    )

    print("Looking for the best assembly plan...")
    t0 = time.time()
    quote = assembly_station.get_quote(sequence, with_assembly_plan=True)
    assembly_plan_report = quote.to_assembly_plan_report()
    assembly_plan_report.write_full_report(
        f"{output_directory}/oligo_assembly_plan_{seq_id}.zip"
    )
    original_sequence = assembly_plan_report.plan.sequence
    # Then get the sequence
    rows = []
    for oligo in assembly_plan_report.plan.assembly_plan:
        # If this was chosen then choose it
        if oligo.accepted:
            rows.append([oligo.id, oligo.sequence, original_sequence])
    return rows


def codon_optimize(protein_sequence: str, min_gc=0.3, max_gc=0.7):
    """Codon optimize the protein sequence using DNA chisel: https://github.com/Edinburgh-Genome-Foundry/DnaChisel"""
    seq = dnachisel.reverse_translate(protein_sequence)
    problem = dnachisel.DnaOptimizationProblem(
        sequence=seq,
        constraints=[
            AvoidPattern("BsaI_site"),  # type: ignore
            EnforceGCContent(mini=min_gc, maxi=max_gc, window=50),  # type: ignore
            EnforceTranslation(location=(0, len(seq))),  # type: ignore
            AvoidStopCodons(location=(0, len(seq) - 3)),  # type: ignore Let there be stop codons in the last bit
        ],
        objectives=[CodonOptimize(species="e_coli", location=(0, len(seq)))],  # type: ignore
    )
    # SOLVE THE CONSTRAINTS, OPTIMIZE WITH RESPECT TO THE OBJECTIVE
    problem.resolve_constraints()
    problem.optimize()

    # PRINT SUMMARIES TO CHECK THAT CONSTRAINTS PASS
    print(problem.constraints_text_summary())
    print(problem.objectives_text_summary())

    # GET THE FINAL SEQUENCE (AS STRING OR ANNOTATED BIOPYTHON RECORDS)
    final_sequence = problem.sequence  # string
    final_record = problem.to_record(with_sequence_edits=True)
    print(protein_sequence)
    print(final_sequence)
    return final_sequence
